@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.030';
+our $VERSION = '0.031';
 
 use EV;
 use Time::HiRes;
@@ -89,8 +89,8 @@ sub new {
 
     # wrapping target selection here allows all client share same list
     # and more importantly to share index within the list
-    $args->{_target}  = _wrap_host($args->{host})             if $args->{host};
-    $args->{_backoff} = _wrap_backoff($args->{backoff_delay}) if $args->{backoff_delay};
+    $args->{_target}  = _wrap_host(delete $args->{host})             if $args->{host};
+    $args->{_backoff} = _wrap_backoff(delete $args->{backoff_delay}) if $args->{backoff_delay};
 
     my %storage;
     my $self = bless {
@@ -128,8 +128,9 @@ sub request {
         if exists $self->{connections}{$conn_id};
 
     my $pool_args = $self->{pool_args};
-    do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme request_timeout
-                                                                               connect_timeout drain_timeout/);
+    do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme
+                                                                               request_timeout connect_timeout
+                                                                               drain_timeout lifetime_timeout/);
     if ($request->{host}) {
         $request->{_target} = _wrap_host($request->{host});
     } elsif ($pool_args->{_target}) {
@@ -232,16 +233,22 @@ sub yahc_reinit_conn {
     return unless defined $args && ref($args) eq 'HASH';
 
     my $request = $conn->{request};
+    $request->{_target}  = _wrap_host(delete $args->{host})             if $args->{host};
+    $request->{_backoff} = _wrap_backoff(delete $args->{backoff_delay}) if $args->{backoff_delay};
     do { $request->{$_} = $args->{$_} if $args->{$_} } foreach (keys %$args);
-    $request->{_target}  = _wrap_host($args->{host})             if exists $args->{host};
-    $request->{_backoff} = _wrap_backoff($args->{backoff_delay}) if exists $args->{backoff_delay};
 }
 
 sub yahc_retry_conn {
-    my ($conn) = @_;
+    my ($conn, $args) = @_;
     die "YAHC: cannot retry completed connection\n"
         if $conn->{state} >= YAHC::State::COMPLETED();
-    $conn->{state} = YAHC::State::INITIALIZED() if yahc_conn_attempts_left($conn);
+    return unless yahc_conn_attempts_left($conn) > 0;
+
+    $conn->{state} = YAHC::State::INITIALIZED();
+    return unless defined $args && ref($args) eq 'HASH';
+
+    $conn->{request}{_backoff} = _wrap_backoff($args->{backoff_delay})
+        if $args->{backoff_delay};
 }
 
 sub yahc_conn_last_error {
@@ -415,9 +422,14 @@ sub _init_helper {
     my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
 
     my $request = $conn->{request};
-    _set_request_timer($self, $conn_id)    if $request->{request_timeout};
-    _set_connection_timer($self, $conn_id) if $request->{connect_timeout};
-    _set_drain_timer($self, $conn_id)      if $request->{drain_timeout};
+
+    $self->{loop}->now_update; # update time for timers
+    _set_until_state_timer($self, $conn_id, 'request_timeout', YAHC::State::USER_ACTION(), YAHC::Error::TIMEOUT() | YAHC::Error::REQUEST_TIMEOUT())
+        if $request->{request_timeout};
+    _set_until_state_timer($self, $conn_id, 'connect_timeout', YAHC::State::CONNECTED(),   YAHC::Error::TIMEOUT() | YAHC::Error::CONNECT_TIMEOUT())
+        if $request->{connect_timeout};
+    _set_until_state_timer($self, $conn_id, 'drain_timeout',   YAHC::State::READING(),     YAHC::Error::TIMEOUT() | YAHC::Error::DRAIN_TIMEOUT())
+        if $request->{drain_timeout};
 
     eval {
         my ($host, $ip, $port, $scheme) = _get_next_target($conn);
@@ -874,10 +886,6 @@ sub yahc_conn_socket_cache_id {
 # Timers
 ################################################################################
 
-sub _set_request_timer    { $_[0]->_set_until_state_timer($_[1], 'request_timeout', YAHC::State::USER_ACTION(), YAHC::Error::TIMEOUT() | YAHC::Error::REQUEST_TIMEOUT()) }
-sub _set_connection_timer { $_[0]->_set_until_state_timer($_[1], 'connect_timeout', YAHC::State::CONNECTED(),   YAHC::Error::TIMEOUT() | YAHC::Error::CONNECT_TIMEOUT()) }
-sub _set_drain_timer      { $_[0]->_set_until_state_timer($_[1], 'drain_timeout',   YAHC::State::READING(),     YAHC::Error::TIMEOUT() | YAHC::Error::DRAIN_TIMEOUT())   }
-
 sub _set_until_state_timer {
     my ($self, $conn_id, $timeout_name, $state, $error_to_report) = @_;
 
@@ -900,7 +908,7 @@ sub _set_until_state_timer {
 
     _register_in_timeline($conn, "setting $timeout_name to %.3fs", $timeout) if exists $conn->{debug_or_timeline};
 
-    $self->{loop}->now_update;
+    # caller should call now_update
     my $w = $watchers->{$timer_name} = $self->{loop}->timer_ns($timeout, 0, $timer_cb);
     $w->priority(2); # set highest priority
     $w->start;
@@ -1241,7 +1249,7 @@ When a connection enters this state C<callback> CodeRef is called:
                 $strerror       # string representation of error
             ) = @_;
 
-            # Note that fields in $conn->{response} are not set 
+            # Note that fields in $conn->{response} are not reliable
             # if $error != YAHC::Error::NO_ERROR()
 
             # HTTP response is stored in $conn->{response}.
@@ -1252,18 +1260,20 @@ When a connection enters this state C<callback> CodeRef is called:
         }
     });
 
-If there was no IO error C<yahc_conn_response> return value is a C<HashRef>
-representing a response. It contains the following key-value pairs.
+If there was no IO error C<yahc_conn_response> return C<HashRef> representing
+response. It contains the following key-value pairs.
 
     proto         => :Str
     status        => :StatusCode
     body          => :Str
     head          => :HashRef
 
-In case of error or non-200 HTTP response C<yahc_retry_conn> or
+In case of a error or non-200 HTTP response C<yahc_retry_conn> or
 C<yahc_reinit_conn> may be called to give the request more chances to complete
 successfully (for example by following redirects or providing new target
-hosts).
+hosts). Also, note that in case of a error data returned by
+C<yahc_conn_response> cannot be trusted. For example, if an IO error happened
+during receiving HTTP body headers would state 200 response code.
 
 In some cases connection cannot be retried anymore and callback is
 called for information purposes only. This case can be distinguished by
@@ -1602,7 +1612,10 @@ connection is in 'USER ACTION' state.
 =head2 yahc_retry_conn
 
 Retries given connection. C<yahc_retry_conn> should be called only if
-C<yahc_conn_attempts_left> returns positive value. Otherwise, it exits silently.
+C<yahc_conn_attempts_left> returns positive value. Otherwise, it exits
+silently. The function accepts HashRef as second argument. By passing it one
+can change C<backoff_delay> parameter. See docs for C<request> for more details
+about C<backoff_delay>.
 
 Intended usage is to retry transient failures or to try different host:
 
